#ifndef SPLINESGRID_H
#define SPLINESGRID_H

// ----------------------------------------------------------------------
//
// Grid.h
//
// Classes embodying the Grid concept for Spline
//
// The generic d-dimensional Grid<d> class template
// Specialization Grid<1> which allows multiple regions
//
// ----------------------------------------------------------------------

#include "splines/Splines_exception.h"
#include <vector>
#include <iostream>
#include <cassert>

namespace splines {

// The following template will be defined in this header:

template <unsigned int d> class Grid;

// The following specializations will be defined in this header:

template <> class Grid<1>;

// Forward declarations of classes used by implementation of Spline<d>

struct GridSearchAnswer;



// ========
// Grid <1> 
// ========

template <>
class Grid<1> {

public:

  // Constructors and assignment

  Grid ( double lowerLimit, 		// A grid with uniform spacing
	 double upperLimit,		// on the HALF-OPEN interval [a,b)
	 unsigned int nGridPoints );	// Grid<1>(0, 100, 100) would be
					// {0, 1, 2, 3, ..., 98, 99}



  template <class iter> inline
  Grid ( iter begin, iter end );	// A grid with non-uniform spacing
					// iter must support:
					// operator++()
					// double operator*()
					// operator != to compare to end

  inline Grid();			// An empty Grid is unsuitable
					// for constructing a Spline

  Grid (const Grid<1> & g);
  inline Grid<1> & operator= (const Grid<1> & g);
  void swap (Grid<1> & g);
  ~Grid() { }

  // Pasting two Grid<1>'s together:

  Grid<1> & operator += (const Grid<1> & g);
  void push_back   (double p); 

  Grid<1> operator +  (const Grid<1> & g) {
    Grid<1> temp(*this);
    return (temp += g);
  }

  // General boilerplate

  bool operator== (Grid<1> const & rhs) const;

  bool operator!= (Grid<1> const & rhs) const { return !(operator==(rhs)); }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // Methods below this line are for use by the Spline class.
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

  // Finding a position in the Grid

  bool search (double p, GridSearchAnswer & answer) const;
  bool search_and_push_back 
  		(double p, std::vector<GridSearchAnswer> & answer) const;

  // Access to coordinate values

  double operator[] (unsigned int p) const;
  std::vector<double> points () const;

  // Information about the Grid

  unsigned int nPoints () const {
    return pointCount;
  }

  std::vector<int> shape () const {
    std::vector<int> s;
    s.push_back(pointCount);
    return s;
  }

  // shifting the origin
  
  void shift(double delta);

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

  std::ostream & put (std::ostream & os) const;
  std::istream & get (std::istream & is);

private:

  double upperBoundary;
  double penultimatePoint;
  double penultimateK;
  int lastIndex;
  unsigned int pointCount;

  std::vector<double> x;	// starting point coordinate for each region
  std::vector<double> h;	// interpoint distance in region point
  std::vector<double> k;	// 1/distance to next point
  std::vector<int> index;	// index in table for starting point of region

}; // Grid<1>

inline
std::ostream& operator<< (std::ostream & os, const Grid<1> & g) {
  return g.put(os);
}



// ========
// Grid <d> 
// ========

template <unsigned int d>
class Grid {

public:

  // Constructors and assignment

  Grid (const Grid<d> & other);

  Grid() { }				// A default Grid is unsuitable
					// for constructing a Spline

  Grid<d> & operator= (const Grid<d> & other);
  void swap (Grid<d> & other);
  ~Grid() { }

  // Outer multiplication to form a higher-dimensional Grid

  template <unsigned int n>
  Grid<d+n> & operator*= (const Grid<n> & other);

  template <unsigned int n>
  Grid<d+n> operator* (const Grid<n> & other);

  // General boilerplate

  bool operator== (Grid<2> const & rhs) const { return g==rhs.g; }

  bool operator!= (Grid<2> const & rhs) const { return !(operator==(rhs)); }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // Methods below this line are for use by the Spline class.
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

  // Finding a position in the Grid

  bool search (const std::vector<double> & v, 
   		     std::vector<GridSearchAnswer> & answer) const;

  // Access to coordinate values

  double coordinate ( int dimension, int index ) {
    return g[dimension][index]; 
  }

  std::vector<double> points ( int dimension ) {
    return g[dimension].points();
  }

  // Information about the Grid

  std::vector<int> shape () const {
    std::vector<int> s;
    for ( int i = 0; i < g.size(); ++i ) s.push_back(g[i].nPoints());
    return s;
  }

  int nPoints() const {
    std::vector<int> s = shape();
    int n = 1;
    for (int i=0; i<s.size(); ++i) { n *= s[i]; }
    return n;
  } 

  std::ostream & put (std::ostream & os) const;
  std::istream & get (std::istream & is);

#ifdef TODO 
	- add access methods as needed by Spline classes
#endif

private:

  std::vector < Grid<1> > g;

}; // Grid<d>



// ========================
// Implementation templates 
// ========================

// ========
// Grid <1> 
// ========

Grid<1>::Grid () : lastIndex(-1), pointCount(0) { }

Grid<1> &
Grid<1>::operator= (const Grid<1> & g) {
  Grid<1> temp(g);
  swap(temp);
  return *this;
}


template <class iter> 
Grid<1>::Grid ( iter begin, iter end ) {
  pointCount = 0;
  lastIndex = -1;
  while ( begin != end ) {
    double newpoint = *begin;
    push_back(newpoint);
    ++begin;
  }
}



// ========
// Grid <d>
// ========

template <unsigned int d>
Grid<d>::Grid (const Grid<d> & other) : g(other.g) { }

template <unsigned int d>
void 
Grid<d>::swap (Grid<d> & other) {
  g.swap(other.g);
}

template <unsigned int d>
Grid<d> & Grid<d>::operator= (const Grid<d> & other) {
  Grid<d> temp(other);
  swap(temp);
  return *this;
}

template <unsigned int d>
bool
Grid<d>::search (const std::vector<double> & v, 
   		       std::vector<GridSearchAnswer> & answer) const {
  assert (d == v.size());
  bool inRange = true;
  std::vector < Grid<1> >::const_iterator i(g.begin());
  std::vector < Grid<1> >::const_iterator end(g.end());
  std::vector<double>::const_iterator p = v.begin();
  while ( i != end ) {
    inRange &= i->search_and_push_back(*p++,answer);
    ++i;
  }
  return inRange;
}

template <unsigned int d>
template <unsigned int n>
Grid<d+n> &
Grid<d>::operator *= (const Grid<n> & other) {
  copy ( other.g.begin(), other.g.end(), back_inserter(g) );    
  return *this;
}

template <unsigned int d>
template <unsigned int n>
Grid<d+n> 
Grid<d>::operator * (const Grid<n> & other) {
  Grid<d> temp(*this);
  return (temp *= other);
}

template<unsigned int d>
std::ostream &
Grid<d>::put(std::ostream & os) const {
  for (unsigned int i=0; i != d; ++i) {
    g[i].put(os);
  }
  return os;
}

template <unsigned int d>
std::ostream& operator<< (std::ostream& os, Grid<d> const & g) {
  return g.put(os);
}

template <unsigned int d>
std::istream& operator>> (std::istream& is, Grid<d>  & g) {
  return g.get(is);
}

} // end of namespace splines



#endif  /* SPLINESGRID_H */
